import {inject, singleton} from 'tsyringe'
import {models} from '@tonality/library-common'
import Uuid, {UuidBuffer} from '@tonality/library-uuid';
import {PrismaClient} from '@prisma/client'
import {
	InsufficientDataError,
	OwnershipError,
	RingtoneNotFoundError,
	UpdateDeletedRingtoneError,
} from './response';

export default interface RingtoneService {
	get(id: Uuid, requesterUserSub?: string): Promise<models.Ringtone>
	browseByComposer(composerUserSub: string, skip?: number, take?: number, requesterUserSub?: string): Promise<models.Ringtone[]>
	browse(skip?: number, take?: number, requesterUserSub?: string): Promise<models.Ringtone[]>
	search(q?: string, requesterUserSub?: string): Promise<models.Ringtone[]>
	create(data: Partial<models.Ringtone>, requesterUserSub: string): Promise<models.Ringtone>
	update(data: Partial<models.Ringtone>, requesterUserSub: string): Promise<models.Ringtone>
	softDelete(id: Uuid, requesterUserSub: string): Promise<models.Ringtone>
	undoDelete(id: Uuid, requesterUserSub: string): Promise<models.Ringtone>
	hardDelete(id: Uuid, requesterUserSub: string): Promise<void>
}

const serializeRingtone = d => {
	return ({
		...d,
		id: Uuid.from(d.id).toString(),
	})
}

@singleton()
export class RingtoneServiceImpl implements RingtoneService {
	constructor(
		@inject('PrismaClient')
		private readonly prismaClient: PrismaClient,
	) {}

	async browseByComposer(composerUserSub: string, skip: number = 0, take: number = 10, requesterUserSub?: string): Promise<models.Ringtone[]> {
		let rawData

		if (typeof requesterUserSub === 'string' && composerUserSub === requesterUserSub) {
			rawData = await this.prismaClient.ringtone.findMany({
				where: {
					composerUserSub: {
						equals: composerUserSub,
					},
				},
				skip,
				take,
			})
		} else {
			rawData = await this.prismaClient.ringtone.findMany({
				where: {
					composerUserSub: {
						equals: composerUserSub,
					},
					deletedAt: {
						equals: null,
					},
				},
				skip,
				take,
			})
		}

		return rawData.map(serializeRingtone)
	}

	async get(id: Uuid, requesterUserSub?: string): Promise<models.Ringtone> {
		const ringtone = await this.prismaClient.ringtone.findUnique({
			where: {
				id: (id as UuidBuffer).toBuffer(),
			}
		})

		if (!ringtone) {
			throw new RingtoneNotFoundError('Ringtone not found!')
		}

		if (ringtone.deletedAt instanceof Date) {
			if (typeof requesterUserSub !== 'string') {
				throw new RingtoneNotFoundError('Ringtone not found!')
			}
			if (requesterUserSub !== ringtone.composerUserSub) {
				throw new RingtoneNotFoundError('Ringtone not found!')
			}
		}

		return serializeRingtone(ringtone)
	}

	async browse(skip: number = 0, take: number = 10, requesterUserSub?: string): Promise<models.Ringtone[]> {
		let rawData
		if (typeof requesterUserSub === 'string') {
			rawData = await this.prismaClient.ringtone.findMany({
				where: {
					OR: [
						{
							deletedAt: {
								equals: null,
							},
						},
						{
							composerUserSub: {
								equals: requesterUserSub,
							},
						}
					],
				},
				skip,
				take,
			})
		} else {
			rawData = await this.prismaClient.ringtone.findMany({
				where: {
					deletedAt: {
						equals: null,
					},
				},
				skip,
				take,
			})
		}

		return rawData.map(serializeRingtone)
	}

	async search(q: string = '', requesterUserSub?: string): Promise<models.Ringtone[]> {
		let rawData
		if (typeof requesterUserSub === 'string') {
			rawData = this.prismaClient.ringtone.findMany({
				where: {
					OR: [
						{
							deletedAt: {
								equals: null,
							},
						},
						{
							composerUserSub: {
								equals: requesterUserSub,
							},
						},
					],
					name: {
						contains: q,
					},
				},
			})
		} else {
			rawData = this.prismaClient.ringtone.findMany({
				where: {
					name: {
						contains: q,
					},
					deletedAt: {
						equals: null,
					},
				},
			})
		}
		return rawData.map(serializeRingtone)
	}

	async create(data: Partial<models.Ringtone>, requesterUserSub: string): Promise<models.Ringtone> {
		const { createdAt, updatedAt, deletedAt, name, composerUserSub, ...safeData } = data
		if (typeof composerUserSub !== 'string') {
			throw new InsufficientDataError('Composer user ID is required.')
		}
		if (composerUserSub !== requesterUserSub) {
			throw new OwnershipError('Creation is only possible with own ringtones.')
		}
		const rawData = await this.prismaClient.ringtone.create({
			data: {
				...safeData,
				id: Uuid.new().toBuffer(),
				composerUserSub: requesterUserSub,
				name: name as string,
			},
		})
		return serializeRingtone(rawData)
	}

	async update(data: Partial<models.Ringtone>, requesterUserSub: string): Promise<models.Ringtone> {
		const { createdAt, updatedAt, deletedAt, id, ...safeData } = data
		const idBuffer = (id as UuidBuffer).toBuffer()
		const checkData = await this.prismaClient.ringtone.findUnique({
			where: {
				id: idBuffer,
			},
		})

		if (!checkData) {
			throw new RingtoneNotFoundError('Ringtone does not exist.')
		}

		if (checkData.deletedAt instanceof Date) {
			throw new UpdateDeletedRingtoneError('Ringtone has been deleted.')
		}

		if (checkData.composerUserSub !== requesterUserSub) {
			throw new OwnershipError('Update is only possible with own ringtones.')
		}

		const rawData = await this.prismaClient.ringtone.update({
			where: {
				id: idBuffer,
			},
			data: {
				...safeData,
				updatedAt: new Date(),
			},
		})
		return serializeRingtone(rawData)
	}

	async softDelete(id: Uuid, requesterUserSub: string): Promise<models.Ringtone> {
		const idBuffer = (id as UuidBuffer).toBuffer()
		const checkData = await this.prismaClient.ringtone.findUnique({
			where: {
				id: idBuffer,
			},
		})
		if (checkData.composerUserSub !== requesterUserSub) {
			throw new OwnershipError('Soft deletion is only possible with own ringtones.')
		}
		const rawData = await this.prismaClient.ringtone.update({
			where: {
				id: idBuffer,
			},
			data: {
				deletedAt: new Date(),
			},
		})
		return serializeRingtone(rawData)
	}

	async undoDelete(id: Uuid, requesterUserSub: string): Promise<models.Ringtone> {
		const idBuffer = (id as UuidBuffer).toBuffer()
		const checkData = await this.prismaClient.ringtone.findUnique({
			where: {
				id: idBuffer,
			},
		})
		if (checkData.composerUserSub !== requesterUserSub) {
			throw new OwnershipError('Undo deletion is only possible with own ringtones.')
		}
		const rawData = this.prismaClient.ringtone.update({
			where: {
				id: (id as UuidBuffer).toBuffer(),
			},
			data: {
				deletedAt: null,
			},
		})
		return serializeRingtone(rawData)
	}

	async hardDelete(id: Uuid, requesterUserSub: string): Promise<void> {
		const idBuffer = (id as UuidBuffer).toBuffer()
		const checkData = await this.prismaClient.ringtone.findUnique({
			where: {
				id: idBuffer,
			},
		})

		if (!checkData) {
			throw new RingtoneNotFoundError('Ringtone does not exist before deletion.')
		}

		if (checkData.composerUserSub !== requesterUserSub) {
			throw new OwnershipError('Hard deletion is only possible with own ringtones.')
		}

		await this.prismaClient.ringtone.delete({
			where: {
				id: idBuffer,
			},
		})
	}
}
